from setuptools import setup, find_packages
setup(
  name = 'melee',
  packages = ['melee'],
  version = '0.0.38',
  description = 'Open API for making your own Smash Bros: Melee AI',
  author = 'AltF4',
  author_email = 'altf4@phx2600.org',
  url = 'https://github.com/altf4/libmelee',
  download_url = 'https://github.com/altf4/libmelee/tarball/0.0.38',
  keywords = ['dolphin', 'AI', 'video games'],
  classifiers = [],
  license = "GPLv3",
  include_package_data=True
)
